@@ -1,12 +1,7 @@
-#Progress percent
-from concurrent.futures import thread
-from multiprocessing import current_process
 import sys, hashlib, logging, shutil, sqlite3, threading, re, fnmatch, os, datetime, json, atexit, subprocess
-from exceptions import FylexError, InvalidPathError, PermissionDeniedError
 from typing import Optional
 
-input_lock = threading.Lock()
-state_lock = threading.Lock()
+thread_lock = threading.Lock()
 
 from pathlib import Path
 from typing import List, Union
@@ -46,13 +41,13 @@ state = FylexState()
 
 # --- Process IDs ----
 def progress():
-    with state_lock:
+    with thread_lock:
         return int(100*state.total_memory_operated/state.total_memory_operation) if state.total_memory_operation else 100
 
 def json_writer():
     json_dir = Path("json")
     json_dir.mkdir(exist_ok=True)
-    with state_lock:
+    with thread_lock:
         file_path = json_dir / f"{state.current_process}.jsonl"
         if not file_path.exists():
             with open(file_path, "w", encoding="utf-8") as f:
@@ -72,7 +67,7 @@ def json_writer():
     return 0
 
 def finalize_json():
-    with state_lock:
+    with thread_lock:
         process_id = state.current_process
     jsonl_path = Path("json") / f"{process_id}.jsonl"
     json_path = Path("json") / f"{process_id}.json"
@@ -102,14 +97,33 @@ def finalize_json():
     return json_path
 
 # --- Raw copier ---
-def get_optimal_buffer_size(src: Path | str):
-    file_size = src.stat().st_size
-    if file_size < 100 * 1024 * 1024:  # < 100 MB
-        return 1 * 1024 * 1024         # 1 MB
-    elif file_size < 10 * 1024 * 1024 * 1024:  # < 10 GB
-        return 8 * 1024 * 1024         # 8 MB
+def get_optimal_buffer_size(src: Path | str) -> int:
+    """
+    Choose an optimal buffer size for copying a file based on its size.
+    Falls back safely if file does not exist.
+    """
+    src = Path(src)
+
+    try:
+        if src.is_file():
+            file_size = src.stat().st_size
+        else:
+            # Not a file → just use safe default
+            return 8 * 1024 * 1024  # 8 MB
+    except FileNotFoundError:
+        # File missing (e.g., was moved/deleted) → safe default
+        return 8 * 1024 * 1024
+    except OSError as e:
+        logging.warning(f"[BUFFER] Could not stat {src}: {e}, using default")
+        return 8 * 1024 * 1024
+
+    if file_size < 100 * 1024 * 1024:         # < 100 MB
+        return 1 * 1024 * 1024                # 1 MB
+    elif file_size < 10 * 1024 * 1024 * 1024: # < 10 GB
+        return 8 * 1024 * 1024                # 8 MB
     else:
-        return 32 * 1024 * 1024        # 32 MB
+        return 32 * 1024 * 1024               # 32 MB
+
 
 def has_cmd(cmd: str) -> bool:
     """Check if a system command exists in PATH."""
@@ -125,39 +139,52 @@ def copier(src: Path | str, dest: Path | str, buffer_size: int,
     size = src.stat().st_size
 
     if not dry_run:
-        # Copy file contents
         with open(src, "rb") as fsrc, open(tmp_dest, "wb") as fdest:
+            in_fd, out_fd = fsrc.fileno(), fdest.fileno()
+            remaining = size
+            offset = 0
+
             if hasattr(os, "copy_file_range"):
-                offset = 0
-                while offset < size:
-                    n = os.copy_file_range(fsrc.fileno(), fdest.fileno(), size - offset)
-                    if n == 0:
-                        raise RuntimeError("copy_file_range() returned 0 before completion.")
-                    offset += n
+                try:
+                    while remaining > 0:
+                        to_copy = min(buffer_size, remaining)
+                        try:
+                            # Modern signature: (fd_in, fd_out, count)
+                            n = os.copy_file_range(in_fd, out_fd, to_copy)
+                        except TypeError:
+                            # Older signature: (fd_in, fd_out, offset_in, offset_out, count)
+                            n = os.copy_file_range(in_fd, out_fd, offset, None, to_copy)
+                        if not n:
+                            raise RuntimeError("copy_file_range() returned 0 before completion")
+                        remaining -= n
+                        offset += n
+                except Exception as e:
+                    logging.debug(f"[COPY] copy_file_range failed, falling back: {e}")
+                    fsrc.seek(0)
+                    fdest.seek(0)
+                    shutil.copyfileobj(fsrc, fdest, length=buffer_size)
+
             elif hasattr(os, "sendfile"):
-                offset = 0
-                while offset < size:
-                    sent = os.sendfile(fdest.fileno(), fsrc.fileno(), offset, size - offset)
+                while remaining > 0:
+                    sent = os.sendfile(out_fd, in_fd, offset, remaining)
                     if sent == 0:
-                        raise RuntimeError("sendfile() returned 0 before completion.")
+                        raise RuntimeError("sendfile() returned 0 before completion")
                     offset += sent
+                    remaining -= sent
             else:
                 shutil.copyfileobj(fsrc, fdest, length=buffer_size)
+
             fdest.flush()
             os.fsync(fdest.fileno())
 
         if preserve_meta:
             shutil.copystat(src, tmp_dest, follow_symlinks=False)
-
-            # Preserve ownership
             if hasattr(os, "chown"):
                 try:
                     st = os.stat(src, follow_symlinks=False)
                     os.chown(tmp_dest, st.st_uid, st.st_gid, follow_symlinks=False)
                 except (PermissionError, AttributeError, NotImplementedError):
                     pass
-
-            # Preserve xattrs
             if has_cmd("getfattr") and has_cmd("setfattr"):
                 try:
                     xattr_out = subprocess.run(
@@ -176,6 +203,8 @@ def copier(src: Path | str, dest: Path | str, buffer_size: int,
                     logging.warning(f"[XATTR] Could not preserve xattrs for {src}: {e}")
             else:
                 logging.warning("Skipping xattr preservation (requires `attr` package).")
+
+            # Preserve ACLs
             if has_cmd("getfacl") and has_cmd("setfacl"):
                 try:
                     acl_out = subprocess.run(
@@ -191,17 +220,19 @@ def copier(src: Path | str, dest: Path | str, buffer_size: int,
                     logging.warning(f"[ACL] Could not preserve ACLs for {src}: {e}")
             else:
                 logging.warning("Skipping ACL preservation (requires `acl` package).")
-        Path(tmp_dest).replace(actual_dest)
-    process_json_writer(src, actual_dest, "copy", algo, dry_run)
 
+        safe_replace(tmp_dest, actual_dest)
+    process_json_writer(src, actual_dest, "copy", algo, dry_run)
 
 # -------- Updater --------
 def func_route_updater(func_name: str):
-    with state_lock:
+    with thread_lock:
         if not state.func_route:
+            # fresh log file only on first use
             with open('fylex.log', 'w') as f:
                 pass
-        state.func_route.append(func_name)
+        # overwrite instead of append, so [0] is always the current func
+        state.func_route = [func_name]
     return True
 
 # -------- Logger Setup --------
@@ -219,22 +250,33 @@ class InfoToLogger:
     def flush(self): pass
 
 def safe_logging(verbose=True):
-    for h in logging.root.handlers[:]:
-        logging.root.removeHandler(h)
-    logging.basicConfig(
-        level=logging.INFO,
-        format="%(asctime)s | %(levelname)s | %(message)s",
-        handlers=[
-            logging.FileHandler("fylex.log", mode="a", encoding="utf-8"),
-            logging.StreamHandler(sys.__stdout__) if verbose else logging.NullHandler(),
-        ],
-    )
-    sys.stdout = InfoToLogger(verbose) 
+    root = logging.getLogger()
+    root.setLevel(logging.INFO)
+
+    # Add a FileHandler to write to fylex.log if not already present
+    fpath_abs = os.path.abspath("fylex.log")
+    has_fh = False
+    for h in root.handlers:
+        if isinstance(h, logging.FileHandler) and getattr(h, "baseFilename", None) == fpath_abs:
+            has_fh = True
+            break
+    if not has_fh:
+        fh = logging.FileHandler("fylex.log", mode="a", encoding="utf-8")
+        fh.setFormatter(logging.Formatter("%(asctime)s | %(levelname)s | %(message)s"))
+        root.addHandler(fh)
+
+    # Add a StreamHandler to the real stdout only if verbose and not already added
+    if verbose:
+        has_sh = any(isinstance(h, logging.StreamHandler) and getattr(h, "stream", None) is sys.__stdout__ for h in root.handlers)
+        if not has_sh:
+            sh = logging.StreamHandler(sys.__stdout__)
+            sh.setFormatter(logging.Formatter("%(asctime)s | %(levelname)s | %(message)s"))
+            root.addHandler(sh)
     
 def log_copier(func_name, summary):
     global state
     do_work = False
-    with state_lock:
+    with thread_lock:
         if state.func_route and state.func_route[0] == func_name:
             do_work = True
             process_id = state.current_process
@@ -261,13 +303,13 @@ def log_copier(func_name, summary):
         json_writer()
         finalize_json()
     finally:
-        with state_lock:
+        with thread_lock:
             state = FylexState()
 
 
 # -------- Input Prompt --------
 def ask_user(question: str) -> str:
-    with input_lock:
+    with thread_lock:
         sys.__stdout__.write(question)
         sys.__stdout__.flush()
         return input().strip().lower()
@@ -483,7 +525,7 @@ def list_matching_files(src: Union[Path, str],
                 logging.debug(f"[DEBUG] {filename} -> included")
                 matched_files.append(path.resolve())
                 st = path.stat()
-                with state_lock:
+                with thread_lock:
                     state.total_memory_operation += st.st_size
                 file_sizes.add(st.st_size)
                 file_extns.add(path.suffix.lower())
@@ -492,7 +534,7 @@ def list_matching_files(src: Union[Path, str],
 
 
 # -------- Conflict resolution --------
-def resolve_conflict(src_file: Path, dest_file: Path, resolve: str) -> bool:
+def resolve_conflict(src_file: Path, dest_file: Path, resolve: str) -> Union[bool, Path]:
     resolve = resolve.lower()
     if resolve not in FylexConfig.ON_CONFLICT_MODES:
         raise ValueError(f"Invalid resolve mode: {resolve}")
@@ -502,30 +544,15 @@ def resolve_conflict(src_file: Path, dest_file: Path, resolve: str) -> bool:
 
     if resolve == "skip":
         return False
-
     if resolve == "replace":
         return True
-
     if resolve in ("larger", "smaller"):
-        src_size = src_file.stat().st_size
-        dest_size = dest_file.stat().st_size
-        if resolve == "larger":
-            return src_size > dest_size
-        else:
-            return src_size < dest_size
-
+        return (src_file.stat().st_size > dest_file.stat().st_size) if resolve=="larger" else (src_file.stat().st_size < dest_file.stat().st_size)
     if resolve in ("newer", "older"):
-        src_mtime = src_file.stat().st_mtime
-        dest_mtime = dest_file.stat().st_mtime
-        if resolve == "newer":
-            return src_mtime > dest_mtime
-        else:
-            return src_mtime < dest_mtime
-
+        return (src_file.stat().st_mtime > dest_file.stat().st_mtime) if resolve=="newer" else (src_file.stat().st_mtime < dest_file.stat().st_mtime)
     if resolve == "prompt":
         choice = ask_user(f"File conflict: {dest_file} | Replace with {src_file}? (y/n): ")
         return choice == "y"
-
     if resolve == "rename":
         base, ext = dest_file.stem, dest_file.suffix
         parent = dest_file.parent
@@ -534,10 +561,9 @@ def resolve_conflict(src_file: Path, dest_file: Path, resolve: str) -> bool:
         while new_file.exists():
             counter += 1
             new_file = parent / f"{base} ({counter}){ext}"
-        dest_file = new_file
-        return dest_file
+        return new_file
+    return False
 
-    return False 
 
 def fast_move(src: Path | str, dest: Path | str, algo: str, dry_run: bool) -> bool:
     if not dry_run:
@@ -553,6 +579,15 @@ def fast_move(src: Path | str, dest: Path | str, algo: str, dry_run: bool) -> bo
 
     return True
 
+def safe_replace(src_tmp: Path, dest: Path):
+    try:
+        Path(dest).unlink(missing_ok=True)
+        Path(src_tmp).replace(dest)
+    except Exception as e:
+        logging.error(f"[CRITICAL] Failed to finalize copy {src_tmp} -> {dest}: {e}")
+        raise RuntimeError(f"Failed safe replace: {e}")
+
+
 def create_dirs(path: Path | str, e_ok: bool = True, dry_run: bool = False):
     try:
         if not Path(path).exists():
@@ -579,7 +614,7 @@ def process_json_writer(src: Path | str, dest: Path | str, ops: str, algo: str,
         file_hash = None
         if dest_path and dest_path.is_file() and algo:
             file_hash = get_or_update_file_hash(dest_path, algo)
-        with state_lock:
+        with thread_lock:
             if state.current_process not in state.process_json:
                 state.process_json[state.current_process] = []
 
@@ -599,7 +634,6 @@ def process_json_writer(src: Path | str, dest: Path | str, ops: str, algo: str,
         logging.exception("[process_json_writer] failed to write process entry")
         return False
 
-
 def copy_with_conflict_resolution(mode: str, src: Path | str, dest: Path | str, backup: Path | str, dry_run: bool, algo: str, resolve: str, 
             verbose: bool, preserve_meta: bool, verify: bool, chunk_size: int, trials: int, recurse: bool):
     dest = Path(dest).resolve()
@@ -635,12 +669,12 @@ def copy_with_conflict_resolution(mode: str, src: Path | str, dest: Path | str,
                     logging.error("[ERROR] File transaction failed: Maximum retries(5) exceeded.")
                     return False
             else:
-                with state_lock:
+                with thread_lock:
                     state.total_memory_operated += action.stat().st_size
                     logging.info(f"[PROGRESS: {progress()}%] File copied and hash verified using {algo} successfully.")
                     state.dupe_candidates.setdefault(action.stat().st_size, []).append(get_or_update_file_hash(action, algo, chunk_size))
         else:
-            with state_lock:
+            with thread_lock:
                 state.total_memory_operated += action.stat().st_size
             logging.info(f"[PROGRESS: {progress()}%] File copied using {algo} successfully.")
         return True
@@ -661,16 +695,16 @@ def copy_with_conflict_resolution(mode: str, src: Path | str, dest: Path | str,
                     logging.error("[ERROR] File transaction failed: Maximum retries(5) exceeded.")
                     return False
             else:
-                with state_lock:
+                with thread_lock:
                     state.total_memory_operated += src.stat().st_size
                 logging.info(f"[PROGRESS: {progress()}%] File copied and hash verified using {algo} successfully.")
         else:
-            with state_lock:
+            with thread_lock:
                 state.total_memory_operated += src.stat().st_size
             logging.info(f"[PROGRESS: {progress()}%] File copied using {algo} successfully.")
         return True
     else:
-        with state_lock:
+        with thread_lock:
             state.total_memory_operation -= src.stat().st_size
         logging.info(f"[SKIP] Skipping {src}, conflict resolution = {resolve}")
         return False
@@ -699,82 +733,116 @@ def find_dupe_candidates(dest: Path | str, file_sizes: set[int], file_extns: set
 
 
 # -------- File Processes --------
-def fileops(func_name: str,
-            src: Path | str, dest: Path | str, resolve: str = "rename", algo: str = FylexConfig.DEFAULT_HASH_ALGO, chunk_size: int = FylexConfig.DEFAULT_CHUNK_SIZE, 
-            verbose: bool = True, dry_run: bool = False, summary: Path | str = None, match_regex: str = None, match_names: str = None, match_glob: str = None,
-            exclude_regex: str = None, exclude_names: str = None, exclude_glob: str = None, recursive_check: bool = False, recurse: bool = False, verify: bool = False,
-            has_extension: bool = False, no_create: bool = False, preserve_meta: bool = True, backup: Path | str = "fylex.deprecated") -> bool:
+def fileops(func_name: str, src: Path | str, dest: Path | str, resolve: str = "rename", algo: str = FylexConfig.DEFAULT_HASH_ALGO, 
+            chunk_size: int = FylexConfig.DEFAULT_CHUNK_SIZE, verbose: bool = True, dry_run: bool = False, summary: Path | str = None, 
+            match_regex: str = None, match_names: str = None, match_glob: str = None, exclude_regex: str = None, exclude_names: str = None, 
+            exclude_glob: str = None, recursive_check: bool = False, recurse: bool = False, verify: bool = False, has_extension: bool = False, 
+            no_create: bool = False, preserve_meta: bool = True, backup: Path | str = "fylex.deprecated") -> bool:
     global state
-    if (backup == "fylex.deprecated"):
-        with state_lock:
-            backup = Path(dest) / f"fylex.deprecated/{state.current_process}"
-    if (Path(dest).resolve() == Path(backup).resolve()):
-        raise ValueError("`dest` and `backup` cannot be same directories.")
-    if (not Path(dest).is_dir() and no_create):
-        raise InvalidPathError(dest)
-    if (Path(src).absolute() == Path(dest).absolute()):
-        raise ValueError(f"Source and destination paths cannot be the same.")
-    create_dirs(dest, True, dry_run)
-    create_dirs(Path(dest)/"fylex.tmp", True, dry_run)
+    src = Path(src).resolve()
+    dest = Path(dest).resolve()
+
+    # Basic path checks
+    if not dest.is_dir() and no_create:
+        raise NotADirectoryError(f"The provided path '{dest}' is not a directory and no_create is enabled.")
+
+    if src == dest:
+        raise ValueError("Source and destination paths cannot be the same.")
+    if dest.is_relative_to(src):
+        raise ValueError("Destination cannot be a subdirectory of source.")
+
+    # Ensure destination directories exist
+    create_dirs(dest, e_ok=True, dry_run=dry_run)
+    create_dirs(dest / "fylex.tmp", e_ok=True, dry_run=dry_run)
+
     func_route_updater(func_name)
-    with state_lock:
-        if state.func_route[0] == func_name:
-            while Path(f"json/{state.current_process}.json").is_file() or Path(f"json/{state.current_process}.jsonl").is_file():
-                state.current_process += 1
-            state.process_json[state.current_process] = []
-            state.parameters = {
-                "func_name": func_name,
-                "src": str(src),
-                "dest": str(dest),
-                "resolve": resolve,
-                "algo": algo,
-                "chunk_size": chunk_size,
-                "verbose": verbose,
-                "dry_run": dry_run,
-                "summary": summary,
-                "match_regex": match_regex,
-                "match_names": match_names,
-                "match_glob": match_glob,
-                "exclude_regex": exclude_regex,
-                "exclude_names": exclude_names,
-                "exclude_glob": exclude_glob,
-                "recursive_check": recursive_check,
-                "verify": verify,
-                "has_extension": has_extension,
-                "no_create": no_create,
-                "preserve_meta": preserve_meta,
-                "backup": backup
-            }
+
+    with thread_lock:
+        # Assign a new process ID
+        while Path(f"json/{state.current_process}.json").is_file() or Path(f"json/{state.current_process}.jsonl").is_file():
+            state.current_process += 1
+
+        # Assign backup path after process ID is finalized
+        if backup == "fylex.deprecated":
+            backup = dest / f"fylex.deprecated/{state.current_process}"
+        backup = Path(backup).resolve()
+
+        if backup == dest:
+            raise ValueError("`dest` and `backup` cannot be the same directories.")
+        if backup in dest.parents:
+            raise ValueError("Backup directory cannot be nested inside destination.")
+
+        # Initialize state for this process
+        state.process_json[state.current_process] = []
+        state.parameters = {
+            "func_name": func_name,
+            "src": str(src),
+            "dest": str(dest),
+            "resolve": resolve,
+            "algo": algo,
+            "chunk_size": chunk_size,
+            "verbose": verbose,
+            "dry_run": dry_run,
+            "summary": summary,
+            "match_regex": match_regex,
+            "match_names": match_names,
+            "match_glob": match_glob,
+            "exclude_regex": exclude_regex,
+            "exclude_names": exclude_names,
+            "exclude_glob": exclude_glob,
+            "recursive_check": recursive_check,
+            "verify": verify,
+            "has_extension": has_extension,
+            "no_create": no_create,
+            "preserve_meta": preserve_meta,
+            "backup": str(backup)
+        }
 
     safe_logging(verbose)
-    file_list, file_sizes, file_extns = list_matching_files(src, match_names, match_regex, match_glob, exclude_names, exclude_regex, exclude_glob, recurse)
-    with state_lock:
+
+    # List and filter files
+    file_list, file_sizes, file_extns = list_matching_files(
+        src, match_names, match_regex, match_glob,
+        exclude_names, exclude_regex, exclude_glob,
+        recurse
+    )
+
+    # Find duplicates in destination
+    with thread_lock:
         state.dupe_candidates = find_dupe_candidates(dest, file_sizes, file_extns, recursive_check, has_extension, algo, chunk_size)
 
+    # Process each file
     for file in file_list:
         file_hash = get_or_update_file_hash(file, algo)
-        with state_lock:
+        with thread_lock:
             if file_hash in state.dupe_candidates.get(file.stat().st_size, []):
                 logging.info(f"[DUPLICATE] Duplicate found for: {file}")
                 continue
-        copy_with_conflict_resolution(func_name[4:], file, dest, backup, dry_run, algo, resolve, verbose, preserve_meta, verify, chunk_size, 0, recurse)
+
+        # Copy or move file with conflict resolution
+        copy_with_conflict_resolution(
+            func_name[4:], file, dest, backup, dry_run, algo,
+            resolve, verbose, preserve_meta, verify, chunk_size, trials=0, recurse=recurse
+        )
+
         if func_name == "filemove":
-            with state_lock:
-                target_dir = backup if backup != Path(dest) / f"fylex.deprecated//{state.current_process}" else Path(src) / f"fylex.deprecated//{state.current_process}"
-            create_dirs(target_dir, True, dry_run)
-            target_path = Path(target_dir) / file.name
+            # Move original file to backup
+            create_dirs(backup, e_ok=True, dry_run=dry_run)
+            target_path = backup / file.name
             fast_move(file, target_path, algo, dry_run)
-    with state_lock:
+
+    # Cleanup empty folders
+    with thread_lock:
         logging.info(f"[NOTE] Note down the process ID for future reference: {state.current_process}")
+
+    if backup.exists() and not any(backup.iterdir()):
+        backup.rmdir()
+    temp = dest / "fylex.tmp"
+    if temp.exists() and not any(temp.iterdir()):
+        temp.rmdir()
+
     log_copier(func_name, summary)
-    if backup.exists():
-        if not backup.iterdir():
-            Path(backup).rmdir()
-    temp = Path(dest) / "fylex.tmp"
-    if temp.exists():
-        if not temp.iterdir():
-            Path(temp).rmdir()
-    return True 
+    return True
 
 
 def filecopy(src: Path | str, dest: Path | str, resolve: str = "rename", algo: str = FylexConfig.DEFAULT_HASH_ALGO, chunk_size: int = FylexConfig.DEFAULT_CHUNK_SIZE, 
@@ -796,24 +864,11 @@ def filemove(src: Path | str, dest: Path | str, resolve: str = "rename", algo: s
                    exclude_regex, exclude_names, exclude_glob, recursive_check, recurse,
                    verify, has_extension, no_create, preserve_meta, backup)
 
-
-def undo(p_id: str, algo: str = FylexConfig.DEFAULT_HASH_ALGO, verbose: bool = True, force: bool = False) -> bool:
+def undo(p_id: str, verbose: bool = True, force: bool = False) -> bool:
     global state
     safe_logging(verbose)
     func_name = "undo"
     func_route_updater(func_name)
-    with state_lock:
-        if state.func_route[0] == func_name:
-            while Path(f"json/{state.current_process}.json").is_file():
-                state.current_process += 1
-            state.process_json[state.current_process] = []
-            state.parameters = {
-                "func_name": func_name,
-                "p_id": p_id,
-                "algo": algo,
-                "verbose": verbose,
-                "force": True
-            }
 
     json_path = Path(f"json/{p_id}.json")
     if not json_path.exists():
@@ -823,6 +878,26 @@ def undo(p_id: str, algo: str = FylexConfig.DEFAULT_HASH_ALGO, verbose: bool = T
     with open(json_path, "r", encoding="utf-8") as f:
         data = json.load(f)
 
+    params = data.get("parameters", {})
+    algo = params.get("algo", FylexConfig.DEFAULT_HASH_ALGO)
+    dry_run = params.get("dry_run", False)
+    preserve_meta = params.get("preserve_meta", True)
+
+    with thread_lock:
+        if state.func_route[0] == func_name:
+            while Path(f"json/{state.current_process}.json").is_file():
+                state.current_process += 1
+            state.process_json[state.current_process] = []
+            state.parameters = {
+                "func_name": func_name,
+                "p_id": p_id,
+                "algo": algo,
+                "verbose": verbose,
+                "force": force,
+                "dry_run": dry_run,
+                "preserve_meta": preserve_meta,
+            }
+
     proc_map = data.get("process_json", {})
     try:
         integer_check = int(p_id)
@@ -830,27 +905,32 @@ def undo(p_id: str, algo: str = FylexConfig.DEFAULT_HASH_ALGO, verbose: bool = T
         entries = proc_map.get(str(p_id)) or []
     except ValueError:
         return ValueError("Expected integer for p_id, but received non-integral values")
+
     for entry in reversed(entries):
         if entry.get("dry_run", False):
-            logging.warning("[WARNING] Dry-runs cannot be rolled back using undo.")
+            logging.warning("[UNDO] Skipping dry-run entry (cannot rollback a dry-run).")
             continue
+
         src = Path(entry['src'])
         dest = Path(entry['dest']) if entry['dest'] else None
         op = entry['operation']
+
         try:
             if op == "copy":
                 if dest and dest.exists():
-                    dest.unlink()
+                    if not dry_run:
+                        dest.unlink()
                     logging.info(f"[UNDO] Removed {dest}")
                 else:
-                    logging.warning(f"[WARNING] File at {dest} not found")
+                    logging.warning(f"[UNDO] File at {dest} not found")
             elif op == "move":
                 create_dirs(src.parent, True, False)
                 if dest and dest.exists():
-                    fast_move(dest, src, algo, False)
-                    logging.info(f"[UNDO] File:{dest} has been restored")
+                    if not dry_run:
+                        fast_move(dest, src, algo, False)
+                    logging.info(f"[UNDO] File {dest} has been restored")
                 else:
-                    logging.warning(f"[WARNING] File at {dest} not found")
+                    logging.warning(f"[UNDO] File at {dest} not found")
             elif op == "delete":
                 if src.exists():
                     logging.warning(f"[UNDO] File {src} already exists, cannot restore deleted file")
@@ -859,19 +939,108 @@ def undo(p_id: str, algo: str = FylexConfig.DEFAULT_HASH_ALGO, verbose: bool = T
             elif op == "create":
                 if src.exists():
                     if not any(src.iterdir()):
-                        src.rmdir()
+                        if not dry_run:
+                            src.rmdir()
                         logging.info(f"[UNDO] Removed created directory {src}")
                     else:
                         logging.warning(f"[UNDO] Cannot remove {src} as it has been modified or contains other files")
             else:
-                logging.error(f"[ERROR] Unknown operation performed on file: {dest}")
+                logging.error(f"[UNDO] Unknown operation performed on file: {dest}")
         except Exception as e:
             logging.error(f"[UNDO] Failed to undo {op} for {src}: {e}")
             if not force:
                 return False
+
+    if not entries:
+        logging.info(f"[UNDO] No entries to process in JSON for {p_id}.")
+    else:
+        logging.info(f"[UNDO] Process {p_id} undone successfully")
+
+    state = FylexState()
+    return True
+
+
+def redo(p_id: str, verbose: bool = True, force: bool = False) -> bool:
+    global state
+    safe_logging(verbose)
+    func_name = "redo"
+    func_route_updater(func_name)
+
+    json_path = Path(f"json/{p_id}.json")
+    if not json_path.exists():
+        logging.error(f"[REDO] No record found for process ID {p_id}")
+        return False
+
+    with open(json_path, "r", encoding="utf-8") as f:
+        data = json.load(f)
+
+    params = data.get("parameters", {})
+    algo = params.get("algo", FylexConfig.DEFAULT_HASH_ALGO)
+    dry_run = params.get("dry_run", False)
+    preserve_meta = params.get("preserve_meta", True)
+
+    with thread_lock:
+        if state.func_route[0] == func_name:
+            while Path(f"json/{state.current_process}.json").is_file():
+                state.current_process += 1
+            state.process_json[state.current_process] = []
+            state.parameters = {
+                "func_name": func_name,
+                "p_id": p_id,
+                "algo": algo,
+                "verbose": verbose,
+                "force": force,
+                "dry_run": dry_run,
+                "preserve_meta": preserve_meta,
+            }
+
+    proc_map = data.get("process_json", {})
+    try:
+        integer_check = int(p_id)
+        del integer_check
+        entries = proc_map.get(str(p_id)) or []
+    except ValueError:
+        return ValueError("Expected integer for p_id, but received non-integral values")
+
+    for entry in entries:
+        if entry.get("dry_run", False):
+            logging.info("[REDO] Skipping dry-run entry")
+            continue
+        src = Path(entry['src'])
+        dest = Path(entry['dest']) if entry['dest'] else None
+        op = entry['operation']
+        try:
+            if op == "copy":
+                if force or not (dest and dest.exists()):
+                    copier(src, dest, get_optimal_buffer_size(src),
+                           algo, preserve_meta, mode="copy", dry_run=dry_run)
+                    logging.info(f"[REDO] Re-copied {src} -> {dest}")
+            elif op == "move":
+                if force or not (dest and dest.exists()):
+                    fast_move(src, dest, algo, dry_run)
+                    logging.info(f"[REDO] Re-moved {src} -> {dest}")
+            elif op == "delete":
+                if dest and dest.exists():
+                    if not dry_run:
+                        dest.unlink()
+                    logging.info(f"[REDO] Deleted {dest}")
+                else:
+                    logging.warning(f"[REDO] File at {dest} not found")
+            elif op == "create":
+                if not dry_run:
+                    create_dirs(src, True, dry_run)
+                logging.info(f"[REDO] Re-created directory {src}")
+            else:
+                logging.error(f"[REDO] Unknown operation performed on file: {dest}")
+        except Exception as e:
+            logging.error(f"[REDO] Failed to redo {op} for {src}: {e}")
+            if not force:
+                return False
+
     if not entries:
         logging.info(f"[ERROR] Could not process JSON: {p_id} or it is empty.")
     else:
-        logging.info(f"[UNDO] Process {p_id} undone successfully")
+        logging.info(f"[REDO] Process {p_id} replayed successfully")
+
     state = FylexState()
     return True
